from json import JSONDecodeError
from typing import Type, TypeVar
from pydantic import BaseModel
from langchain_core.messages import BaseMessage
import json

T = TypeVar('T', bound=BaseModel)

def deeper_validate(raw: BaseMessage, format_model: Type[T]) -> T:
    """
    Validate the raw message and return a pydantic model.
    """
    try:
        return format_model.model_validate(raw)
    except Exception as e:
        if 'Invalid json output' not in str(e):
            raise e
    
        _errored_response = str(e)
        first_brace = _errored_response.find('{')
        if first_brace == -1:
            raise e
        
        last_brace = _errored_response.rfind('}')
        if last_brace == -1 and _errored_response[-1] != '}':
            raise e
        
        if first_brace >= last_brace:
            raise e

        json_content = _errored_response[first_brace:last_brace + 1]

        parsed_json = json.loads(json_content)
        return format_model.model_validate(parsed_json)


import os
from typing import Optional, Type, TypeVar, Literal, List, Tuple

from pydantic import BaseModel, SecretStr

from langchain_anthropic import ChatAnthropic
from langchain_core.messages import HumanMessage, SystemMessage, BaseMessage
from langchain_core.output_parsers import JsonOutputParser
import asyncio
from nael_utils.deeper_validate import deeper_validate

T = TypeVar('T', bound=BaseModel)


def call_claude(
    prompt: str,
    format_model: Type[T],
    model_type: Literal["sonnet", "opus"] = "sonnet",
    system_prompt: str = "",
    api_key: Optional[SecretStr] = None,
    additional_messages: Optional[List[BaseMessage]] = None,
    max_tokens: int = 10000,
    thinking_token: int = 0
) -> T:
    
    loop = asyncio.get_event_loop()
    if loop.is_running():
        # fallback for running loops
        import nest_asyncio
        nest_asyncio.apply()

    result, _, _ = loop.run_until_complete(
        acall_claude(
            prompt=prompt, 
            format_model=format_model, 
            model_type=model_type, 
            system_prompt=system_prompt, 
            api_key=api_key, 
            additional_messages=additional_messages, 
            max_tokens=max_tokens, 
            thinking_token=thinking_token
        )   
    ) 
    return result

async def acall_claude(
    prompt: str,
    format_model: Type[T],
    model_type: Literal["sonnet", "opus"] = "sonnet",
    system_prompt: str = "",
    api_key: Optional[SecretStr] = None,
    additional_messages: Optional[List[BaseMessage]] = None,
    max_tokens: int = 10000,
    thinking_token: int = 0
) -> Tuple[T, int, int]:
    """
    Generic function to ask for Claude's inference. Return value will be in format_model
    :param prompt: The prompt to send to Claude
    :param format_model: Pydantic model to structure the response
    :param model_type: Claude model to use
    :param system_prompt: System prompt to set context
    :param max_tokens: Maximum tokens in response
    :param api_key: Anthropic API key
    :return: Parsed response as format_model instance, number of input tokens, number of output tokens
    """

    model = f"claude-{model_type}-4-0"

    if api_key is None:
        _api_key = os.getenv("ANTHROPIC_API_KEY", None)

        if _api_key is None:
            raise ValueError("ANTHROPIC_API_KEY is not set")
        
        api_key = SecretStr(_api_key)

    if thinking_token:
        chat = ChatAnthropic(
            api_key=api_key, 
            model_name=model,
            timeout=None,
            stop=None,
            max_tokens=max_tokens + thinking_token,
            thinking={
                "type": "enabled",
                "budget_tokens": thinking_token,
            },
        )
    else:
        chat = ChatAnthropic(
            api_key=api_key, 
            model_name=model,
            timeout=None,
            stop=None,
            max_tokens=max_tokens,
        )

    parser = JsonOutputParser(pydantic_object=format_model)
    format_instructions = parser.get_format_instructions()
    _prompt = f"Answer the user query.\n{format_instructions}\n{prompt}\n"
    messages: List[BaseMessage] = [
        HumanMessage(
            content=[
                {"type": "text", "text": _prompt},
            ]
        )
    ]

    if system_prompt:
        messages.insert(0, SystemMessage(content=system_prompt))

    if additional_messages:
        messages.extend(additional_messages)

    text_result = await chat.ainvoke(messages)
    result = parser.invoke(text_result)

    return (
        deeper_validate(result, format_model), 
        text_result.usage_metadata["input_tokens"], # type: ignore
        text_result.usage_metadata["output_tokens"] # type: ignore
    )



"""
A collection of utilities for Python.
"""

# Import all public functions and classes
from .decorators import retry, forever
from .anthropic_ext import call_claude
from .langchain_ext import (
    call_gpt_with_prompt_model,
    call_gpt,
)

# Define what gets exported when using "from nael_utils import *"
__all__ = [
    "retry",
    "forever", 
    "call_claude",
    "call_gpt_with_prompt_model",
    "call_gpt",
]

__version__ = "0.1.8"



import os
from typing import Dict, Optional, Type, TypeVar

from langchain_core.messages import BaseMessage, HumanMessage, SystemMessage
from langchain_core.output_parsers import JsonOutputParser
from pydantic import BaseModel, SecretStr
from langchain_openai.chat_models import ChatOpenAI

from nael_utils.deeper_validate import deeper_validate


def call_gpt_with_prompt_model(
    api_key: Optional[str],
    prompt: str,
    pydantic_object: Optional[Type[BaseModel]],
    model: str = "gpt-4.1",
) -> Dict:
    """
    Generic function to ask for GPT's inference. Return value will be in pydantic_object
    :param api_key:
    :param prompt:
    :param pydantic_object:
    :return:
    """
    if api_key is None:
        api_key = os.getenv("OPENAI_API_KEY")
    if api_key is None:
        raise ValueError("OPENAI_API_KEY is not set")
    
    chat = ChatOpenAI(model=model, api_key=SecretStr(api_key))
    parser = JsonOutputParser(pydantic_object=pydantic_object)
    format_instructions = parser.get_format_instructions()
    _prompt = f"Answer the user query.\n{format_instructions}\n{prompt}\n"
    messages = [
        HumanMessage(
            content=[
                {"type": "text", "text": _prompt},
            ]
        )
    ]

    text_result = chat.invoke(messages)
    return parser.invoke(text_result)

T = TypeVar('T', bound=BaseModel)

def call_gpt(
    prompt: str,
    format_model: Type[T],
    model: str = "gpt-4.1",
    system_prompt: str = "",
    additional_messages: list[BaseMessage] = [],
    api_key: Optional[SecretStr] = None,
    max_tokens: Optional[int] = 10000,
) -> T:
    """
    Generic function to ask for GPT's inference. Return value will be in format_model
    :param prompt:
    :param format_model:
    :param model:
    :param system_prompt:
    :param additional_messages:
    :param api_key:
    :return:
    """

    if api_key is None:
        _api_key = os.getenv("OPENAI_API_KEY", None)

        if _api_key is None:
            raise ValueError("OPENAI_API_KEY is not set")
        
        api_key = SecretStr(_api_key)

    chat = ChatOpenAI(model=model, api_key=api_key, max_tokens=max_tokens)
    parser = JsonOutputParser(pydantic_object=format_model)
    format_instructions = parser.get_format_instructions()
    _prompt = f"Answer the user query.\n{format_instructions}\n{prompt}\n"
    messages: list[BaseMessage] = [
        HumanMessage(
            content=[
                {"type": "text", "text": _prompt},
            ]
        ),
        *additional_messages,
    ]

    if system_prompt:
        messages = [
            SystemMessage(content=system_prompt),
            *messages,
        ]

    text_result = chat.invoke(messages)
    result = parser.invoke(text_result)

    return deeper_validate(result, format_model)


from setuptools import setup, find_packages

setup(
    name="nael-utils",
    version="0.1.13",
    description="A collection of utilities for Python.",
    long_description=open('README.md').read(),
    long_description_content_type="text/markdown",
    author="Nael",
    author_email="nael.nathanael71@gmail.com",
    url="https://github.com/Nael-Nathanael/nael-utils",
    packages=find_packages(),
    package_data={
        "nael_utils": ["py.typed"],
    },
    install_requires=[
        "langchain-core",
        "langchain-openai",
        "langchain-anthropic",
        "openai",
        "pydantic",
    ],
    classifiers=[
        "Programming Language :: Python :: 3",
        "License :: OSI Approved :: MIT License",
        "Operating System :: OS Independent",
        "Typing :: Typed",
    ],
    python_requires='>=3.9',
    zip_safe=False,
)
